/* MIFparse.c
 * $Id: MIFparse.c,v 1.1 1995/07/14 15:57:25 connolly Exp $
 */

/* implements... */
#include "MIFparse.h"

/* uses ... */
#include "object.h"
#include "debug.h"
#include <ctype.h>
#include <assert.h>
#include <string.h>

static VOID skip_to_newline PARAMS((HMStream stream, HMGetcProc getch));


static int parse_attributes PARAMS((HMStream stream, HMGetcProc getch,
				     HMBinding* attrs, int nattrs));

static char* parse_string PARAMS((HMStream stream, HMGetcProc getch));

static char* parse_token PARAMS((HMStream stream, HMGetcProc getch, int c));

static char* parse_number PARAMS((HMStream stream, HMGetcProc getch,
				  int* lookahead));



VOID
  MIF_parse(stream, getch, doc, docclass)
HMStream stream;
HMGetcProc *getch;
HMDoc* doc;
CONST HMDoc_Class *docclass;
{
  int c;
  
  while((c = (getch)(stream)) != EOF){
    switch(c){

      /* whitespace */
    case '\n': case '\r': case ' ': case '\t':
      break;

      /* comment */
    case '#':
      skip_to_newline(stream, getch);
      break;

      /* statement */
    case '<':
      {
	static HMBinding attrs[] = { {"A1"}, {"A2"}, {"A3"}, {"A4"} };
	int i;
	char* token = parse_token(stream, getch, EOF);
	int maxattrs = ARRAYSIZE(attrs);
	int nattrs = parse_attributes(stream, getch, attrs, maxattrs);
	int content = (docclass->startTag)(doc, token, attrs, nattrs);

	for(i = 0; i<nattrs; i++)
	  FREE(attrs[i].value);

	if(content != SGML_EMPTY){
	  MIF_parse(stream, getch, doc, docclass);
	  (docclass->endTag)(doc, token);
	}

	FREE(token);
      }
      break;


      /* end of statement */
    case '>':
      return;

    default:
      debug(("error: '%c' unexpected.\n", c));
      break;
    }
  }
}


static VOID
  skip_to_newline(stream, getch)
HMStream stream;
HMGetcProc *getch;
{
  int c;

  do{
    c = (getch)(stream);
  }while( c != EOF & c != '\n');
}


static int
  parse_attributes(stream, getch, attrs, nattrs)
HMStream stream;
HMGetcProc* getch;
HMBinding* attrs;
int nattrs;
{
  int here;
  int c;

  for(here = 0; here < nattrs; here++){
    attrs[here].value = 0;
  }
  
  here = 0;

  c = (getch)(stream);

  while(c != EOF){
    char* value;

    /* end of attributes */
    if(c == '>' || c == '\n'){ /*@@ \n is pretty much a hack. */
      return here;
    }

    /* string */
    else if(c == '`'){
      value = parse_string(stream, getch);
      c = (getch)(stream);
    }
    else if (isdigit(c)){
      value = parse_number(stream, getch, &c);
    }
    else if (isalpha(c)){
      value = parse_token(stream, getch, c);
      c = (getch)(stream);
    }
    else{
      debug(("error: '%c' unexpected in attributes.\n", c));
      c = (getch)(stream);
      continue; /* no value */
    }

    if(here < nattrs){
      attrs[here++].value = value;
    }else{
      FREE(value);
    }
  }

  return here;
}


static char*
  parse_token(stream, getch, c)
HMStream stream;
HMGetcProc* getch;
int c;
{
  int tot = 20;
  int cur = 0;
  char* ret = NEW(char, tot);

  if(c == EOF)
    c = (getch)(stream);

  while(isalnum(c)){
    if(cur + 1 >= tot){
      char* newbuf = NEW(char, tot = tot * 3 / 2);
      if(cur > 0){
	memcpy(newbuf, ret, cur);
	FREE(ret);
      }
      ret = newbuf;
    }

    ret[cur++] = c;
    c = (getch)(stream);
  }

  /* eat lookahead char */

  ret[cur] = 0;
  return ret;
}


#define HEX(c) ( (c)>='0' && (c)<='9' ? (c - '0') : \
		( (c)>= 'a' && (c)<='f' ? (c - 'a' + 10) :\
		 ( (c)>= 'A' && (c)<='F' ? (c - 'A' + 10) :\
		  0)))

static char*
  parse_string(stream, getch)
HMStream stream;
HMGetcProc* getch;
{
  int c;
  int tot = 40;
  int cur = 0;
  char* ret = NEW(char, tot);

  while( (c = (getch)(stream)) != EOF ){
    /* end of string */
    if(c == '\''){
      break;
    }

    /* escape sequence */
    else if(c == '\\'){
      c = (getch)(stream);
      switch(c){

      case EOF:
	ret[cur] = 0; /* null terminate it */
	return ret;

      case 'x':
	{ int c1 = (getch)(stream);
	  int c2 = (getch)(stream);
	  if(c1 == EOF || c2 == EOF){
	    ret[cur] = 0; /* null terminate it */
	    return ret;
	  }
	  c = HEX(c1) * 16 + HEX(c2);
	  break;
	}

      case 't':
	c = '\t';
	break;

      /* case '>':
	 c = '>';
	 break; */

      case 'q':
	c = '\'';
	break;

      case 'Q':
	c = '`';
	break;

      /* case '\':
	 c = '\';
	 break; */
      }
    }

    if(cur + 1 >= tot){
      char* newbuf = NEW(char, tot = tot * 3 / 2);
      if(cur > 0){
	memcpy(newbuf, ret, cur);
	FREE(ret);
      }
      ret = newbuf;
    }

    ret[cur++] = c;
  }

  /* eat lookahead char */

  ret[cur] = 0;
  return ret;
}


static char*
  parse_number(stream, getch, lookahead)
HMStream stream;
HMGetcProc* getch;
int* lookahead;
{
  int tot = 10;
  int cur = 0;
  char* ret = NEW(char, tot);
  int c = *lookahead;
  enum {start, num, units} state = start;
  int decimal = 0;

  if(c == EOF)
    c = (getch)(stream);

  while(c != EOF){
    if(c == '>'){
      break;
    }
    else if(state == start){
      if(isdigit(c)){
	state = num;
	continue;
      }
      else{
	ret[cur++] = '0'; /* start number with a digit */
	state = num;

	if(c == '.')
	  decimal = cur;
	/* and fall through and append c */
      }
    }
    else if(state == num){
      if(! (c == '.' || isdigit(c))){
	decimal = 0;
	state = units;
	continue;
      }
      if(decimal == 0 && c == '.')
	decimal = cur;
      else if(c != '0'){ /* saw non-zero. forget about decimal */
	decimal = 0;
      }
      /* else fall through and append c */
    }

    else{ /* parsing units */
      if(isdigit(c)){
	break; /* stop! start of next number! */
      }
      else if(isspace(c)){
	c = (getch)(stream);
	continue; /* skip whitespace */
      }
      else{
	decimal = 0; /* saw something significant after decimal. */
	if(c == '"'){ /* inch indicator */
	  ret[cur++] = 'i';
	  c = 'n';
	}
      }
      /* else fall through and append c */
    }

    if(cur + 3 >= tot){ /* room for lead '0' char, "in"
			   and terminating null char */
      char* newbuf = NEW(char, tot = tot * 3 / 2);
      if(cur > 0){
	memcpy(newbuf, ret, cur);
	FREE(ret);
      }
      ret = newbuf;
    }

    ret[cur++] = c;
    c = (getch)(stream);
  }


  *lookahead = c;

  if(decimal) /* saw "3.00" or some such. Truncate to "3" */
    ret[decimal] = 0;
  else
    ret[cur] = 0;

  return ret;
}


